from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.color.regex import ColorRegularExpression
from yta_general_utils.color.enums import ColorString


def is_hexadecimal_color(color):
    """
    Check that the 'color' parameter is an hexadecimal
    color.
    """
    return ColorRegularExpression.HEX.parse(color)

def is_string_color(color):
    """
    Check that the 'color' parameter is an string 
    color accepted by our system, whose value is an
    hexadecimal value.
    """
    return ColorString.is_valid(color)

def is_array_or_tuple_without_alpha_normalized(color):
    """
    Check that the 'color' parameter is an array or a
    tuple of 3 elements that are float values between
    0 and 1 (normalized value).
    """
    return is_array_or_tuple_without_alpha and all(PythonValidator.is_instance(c, float) and 0 <= c <= 1 for c in color)

def is_array_or_tuple_with_alpha_normalized(color):
    """
    Check that the 'color' parameter is an array or a
    tuple of 4 elements that are float values between
    0 and 1 (normalized value).
    """
    return is_array_or_tuple_with_alpha and all(PythonValidator.is_instance(c, float) and 0 <= c <= 1 for c in color)

def is_array_or_tuple_without_alpha(color):
    """
    Check that the 'color' parameter is an array or a
    tuple of 3 elements that are int values between 0
    and 255.
    """
    return PythonValidator.is_instance(color, tuple) or PythonValidator.is_instance(color, list) and len(color) == 3 and all(PythonValidator.is_instance(c, int) and 0 <= c <= 255 for c in color)

def is_array_or_tuple_with_alpha(color):
    """
    Check that the 'color' parameter is an array or a
    tuple of 4 elements that are int values between 0
    and 255.
    """
    return PythonValidator.is_instance(color, tuple) or PythonValidator.is_instance(color, list) and len(color) == 4 and all(PythonValidator.is_instance(c, int) and 0 <= c <= 255 for c in color)

def parse_rgb_color(color):
    """
    Parse the provided 'color' as RGB and returns it as
    r,g,b values.
    """
    if is_array_or_tuple_without_alpha_normalized(color):
        return color[0] * 255, color[1] * 255, color[2] * 255
    elif is_array_or_tuple_without_alpha(color):
        return color[0], color[1], color[2]
    else:
        raise Exception(f'The provided "color" parameter is not an RGB color.')

def parse_rgba_color(color):
    """
    Parse the provided 'color' as RGBA and returns it as
    r,g,b,a values.
    """
    if is_array_or_tuple_with_alpha_normalized(color):
        return color[0] * 255, color[1] * 255, color[2] * 255, color[3] * 255
    elif is_array_or_tuple_with_alpha(color):
        return color[0], color[1], color[2], color[3]
    else:
        raise Exception(f'The provided "color" parameter is not an RGBA color.')



from datetime import datetime


def get_rfc_datetime(year: int = 1900, month: int = 1, day: int = 1, hour: int = 0, minute: int = 0):
    """
    Receives a date based on provided parameters and turns it into
    a RFC datetime, returning it.

    TODO: Explain this better, please
    """
    dt = datetime(year, month, day, hour, minute, 0).isoformat() + 'Z'

    return dt



from yta_general_utils.web.scrapper.chrome_scrapper import ChromeScrapper
from yta_general_utils.downloader import get_file
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from typing import Union


def get_facebook_video(url: str, output_filename: Union[str, None] = None):
    """
    Gets the Facebook video (reel) from the provided 'url' (if valid)
    and returns its data or stores it locally as 'output_filename' if
    provided.
    """
    DOWNLOAD_FACEBOOK_VIDEO_URL = 'https://fdownloader.net/en/facebook-reels-downloader'

    scrapper = ChromeScrapper()
    scrapper.go_to_web_and_wait_util_loaded(DOWNLOAD_FACEBOOK_VIDEO_URL)

    # We need to wait until video is shown
    url_input = scrapper.find_element_by_id('s_input')
    url_input.send_keys(url)
    url_input.send_keys(Keys.ENTER)

    # We need to click in the upper left image to activate vid popup
    image_container = scrapper.find_element_by_class_waiting('div', 'image-fb open-popup')
    image = image_container.find_element(By.TAG_NAME, 'img')
    image.click()

    #video_element = scrapper.find_element_by_element_type_waiting('video')
    video_element = scrapper.find_element_by_id_waiting('vid')
    video_source_url = video_element.get_attribute('src')

    return get_file(video_source_url, output_filename)


from yta_general_utils.color.utils import is_hexadecimal_color, parse_rgb_color, parse_rgba_color
from typing import Union


class ColorConverter:
    """
    Class to simplify and encapsulate the functionality related
    to color conversion.
    """
    @staticmethod
    def rgb_to_hex(red, green, blue):
        """
        Returns the provided RGB color as a hex color. The 'red', 'green' and
        'blue' parameters must be between 0 and 255.
        """
        return '#{:02x}{:02x}{:02x}'.format(red, green, blue)
    
    @staticmethod
    def hex_to_rgb(color: str):
        """
        Parse the provided hexadecimal 'color' parameter and
        turn it into an RGB color (returned as r,g,b) or
        raises an Exception if not.
        """
        r, g, b, _ = ColorConverter.hex_to_rgba(color)

        return r, g, b

    @staticmethod
    def hex_to_rgba(color: str):
        if not is_hexadecimal_color(color):
            raise Exception(f'The provided "color" parameter "{str(color)}" is not an hexadecimal color.')
        
        # Hex can start with '0x', '0X' or '#'
        hex = color.lstrip('#').lstrip('0x').lstrip('0X')
        if len(hex) == 8:
            # hex with alpha
            r, g, b, a = (int(hex[i:i+2], 16) for i in (0, 2, 4, 6))
        elif len(hex) == 6:
            # hex without alpha
            r, g, b, a = *(int(hex[i:i+2], 16) for i in (0, 2, 4)), 0
        
        return r, g, b, a

    def rgb_to_hex(color: Union[tuple, list], do_include_alpha: bool = False):
        """
        Parse the provided RGB 'color' parameter and turn it to
        a hexadecimal color if valid or raises an Exception if
        not. The result will be #RRGGBB if 'do_include_alpha' is
        False, or #RRGGBBAA if 'do_include_alpha' is True.
        """
        r, g, b = parse_rgb_color(color)

        hex = "#{:02x}{:02x}{:02x}".format(r, g, b)
        if do_include_alpha:
            hex = "#{:02x}{:02x}{:02x}{:02x}".format(r, g, b, 0)

        return hex
        
    def rgba_to_hex(color: Union[tuple, list], do_include_alpha: bool = False):
        """
        Parse the provided RGBA 'color' parameter and turn it to
        a hexadecimal color if valid or raises an Exception if
        not. The result will be #RRGGBB if 'do_include_alpha' is
        False, or #RRGGBBAA if 'do_include_alpha' is True.
        """
        r, g, b, a = parse_rgba_color(color)

        hex = "#{:02x}{:02x}{:02x}".format(r, g, b)
        if do_include_alpha:
            hex = "#{:02x}{:02x}{:02x}{:02x}".format(r, g, b, a)

        return hex

    def rgba_to_hsl(color: Union[tuple, list]):
        # TODO: Explain
        _, _, _, a = parse_rgba_color(color)
        
        return *ColorConverter.rgb_to_hsl(color), a

    def rgb_to_hsl(color: Union[tuple, list]):
        # TODO: Explain
        r, g, b = parse_rgb_color(color)
        # Normalizamos los valores de r, g, b de 0-255 a 0-1
        r /= 255.0
        g /= 255.0
        b /= 255.0
        
        # Encuentra los valores máximos y mínimos de los componentes RGB
        cmax = max(r, g, b)
        cmin = min(r, g, b)
        delta = cmax - cmin

        # Calcular el tono (H)
        if delta == 0:
            h = 0  # Si no hay diferencia, el tono es indefinido (gris)
        elif cmax == r:
            h = (60 * ((g - b) / delta) + 360) % 360
        elif cmax == g:
            h = (60 * ((b - r) / delta) + 120) % 360
        else:  # cmax == b
            h = (60 * ((r - g) / delta) + 240) % 360
        
        # Calcular la luminosidad (L)
        l = (cmax + cmin) / 2
        
        # Calcular la saturación (S)
        if delta == 0:
            s = 0  # Si no hay diferencia, la saturación es 0 (gris)
        else:
            s = delta / (1 - abs(2 * l - 1)) if l != 0 and l != 1 else delta / (2 - (cmax + cmin))

        # TODO: I saw in some online solutions that they offer
        # the results without decimal figures
        return round(h, 2), round(s * 100, 2), round(l * 100, 2)

    def rgba_to_cymk(color: Union[tuple, list]):
        # TODO: Explain
        # TODO: Is there a way to handle alpha transparency
        # with a cymk (?)
        return ColorConverter.rgb_to_cymk(color)

    def rgb_to_cymk(color: Union[tuple, list]):
        # TODO: Explain
        # It looks like you need to know the color profile before
        # any conversion from RGB or RGBA
        # https://www.reddit.com/r/AdobeIllustrator/comments/17vpbod/different_cmyk_values_for_same_hex_code_which/?rdt=55781#:~:text=A%20hex%20code%20is%20an,information%2C%20like%20a%20colour%20profile.
        r, g, b = parse_rgb_color(color)
        r, g, b = r / 255.0, g / 255.0, b / 255.0

        k = 1 - max(r, g, b)

        if k == 1:
            c = m = y = 0
        else:
            c = (1 - r - k) / (1 - k)
            m = (1 - g - k) / (1 - k)
            y = (1 - b - k) / (1 - k)

        # TODO: I saw in some online solutions that they offer
        # the results without decimal figures
        return round(c * 100, 2), round(m * 100, 2), round(y * 100, 2), round(k * 100, 2)
    



from yta_general_utils.programming.parameter_validator import NumberValidator

import random


def randrangefloat(start: float, end: float, step: float):
    """
    Calculate and return a random float number between the provided
    'start' and 'end' limits using the also provided float 'step'.

    TODO: Is limit included (?) Please, review and, if necessary,
    include it as a parameter.
    """
    # TODO: What about strings that are actually parseable
    # as those numbers (?)
    if not NumberValidator.is_number(start):
        raise Exception('The provided "start" parameter is not a number.')
    
    if not NumberValidator.is_number(end):
        raise Exception('The provided "end" parameter is not a number.')
    
    start = float(start)
    end = float(end)

    # Swap limits if needed
    if end < start:
        start, end = end, start

    # TODO: What if 'start' and 'end' are the same (?)
    return random.choice([round(start + i * step, 4) for i in range(int((end - start) / step) + 1) if start + i * step <= end])

def random_int_between(start: int, end: int, step: int = 1):
    # TODO: What about strings that are actually parseable
    # as those numbers (?)
    if not NumberValidator.is_number(start):
        raise Exception('The provided "start" parameter is not a number.')
    
    if not NumberValidator.is_number(end):
        raise Exception('The provided "end" parameter is not a number.')
    
    start = int(start)
    end = int(end)
    
    # Swap limits if needed
    if end < start:
        start, end = end, start

    # TODO: What if 'start' and 'end' are the same (?)
    return random.randrange(start, end + 1, step)


from yta_general_utils.file.filename import ensure_file_extension, file_has_extension, get_file_extension
from yta_general_utils.file.enums import FileType, AudioFileExtension, VideoFileExtension, ImageFileExtension
from yta_general_utils.temp import create_temp_filename
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.programming.enum import YTAEnum as Enum
from typing import Union


def handle_output_filename(output_filename: Union[None, str], expected_extension: Union[None, AudioFileExtension, VideoFileExtension, ImageFileExtension, FileType] = None):
    """
    Handle the provided 'output_filename' to ensure that it is the
    expected for our system. We could want to write a file or to not
    write it, and we need to make sure the extension is the expected
    one.

    This method should be called within any method that is capable
    to write (mandatory or optional) a filename locally.

    This method will return None when no file writting expected or
    the expected filename when you do want to write.
    """
    if not expected_extension and not output_filename:
        return None
    
    if expected_extension:
        # TODO: Remove this explanation below in a near future
        # When validating types based on Enum items, we can receive
        # the Enum class or an Enum instance. If we receive the Enum class
        # we will be able to choose any of the existing Enum items. If we
        # receive the instance, we will only be able to accept the value
        # of that Enum. If we receive a string and we accept it, we will
        # try to convert it to our accepted Enum classes and, if found,
        # convert to the corresponding instance to work as if the instance
        # was the passed parameter
        if not PythonValidator.is_string(expected_extension) and not PythonValidator.is_class(expected_extension, AudioFileExtension) and not PythonValidator.is_instance(expected_extension, AudioFileExtension) and not PythonValidator.is_class(expected_extension, ImageFileExtension) and not PythonValidator.is_instance(expected_extension, ImageFileExtension) and not PythonValidator.is_class(expected_extension, VideoFileExtension) and not PythonValidator.is_instance(expected_extension, VideoFileExtension):
            raise Exception(f'The provided parameter "expected_extension" is not valid.')
        
        if PythonValidator.is_string(expected_extension):
            expected_extension = Enum.parse_as_enum(expected_extension, [AudioFileExtension, ImageFileExtension, VideoFileExtension])

        # Here 'expected_extension' is a YTAEnum class or instance

    if output_filename is not None and not PythonValidator.is_string(output_filename):
        raise Exception(f'The provided "output_filename" parameter "{str(output_filename)}" is not a string.')
   
    # We don't accept 'output_filename' without extension
    if not expected_extension and not file_has_extension(output_filename):
        raise Exception(f'The provided "output_filename" parameter "{str(output_filename)}" has no valid extension and there is no "expected_extension" parameter provided.')
    
    if expected_extension:
        if PythonValidator.is_a_class(expected_extension):
            # A class, I can accept any of the Enum item values
            default_exception = expected_extension.default().value
            accepted_extensions = expected_extension.get_all_values()
        elif PythonValidator.is_an_instance(expected_extension):
            # An instance, I can only accept that Enum instance value
            default_exception = expected_extension.value
            accepted_extensions = [expected_extension.value]

        if output_filename:
            extension = get_file_extension(output_filename)
            if extension not in accepted_extensions:
                output_filename = ensure_file_extension(output_filename, default_exception)
        else:
            output_filename = create_temp_filename(f'tmp_filename.{default_exception}')

    # TODO: Maybe check if output_filename has no extension to
    # raise Exception if that happens (?)
    return output_filename


from yta_general_utils.image.region import PixelFilterFunction
from yta_general_utils.image.parser import ImageParser
from collections import Counter
from PIL import Image
from typing import Union

import numpy as np


class ColorPicker:
    """
    Class to encapsulate and simplify the functionality related to
    image color detection.
    """
    @staticmethod
    def get_most_common_green_rgb_color(image: Union[str, Image.Image, np.ndarray]):
        """
        Returns the most common (dominant) rgb color in a 
        (r, g, b) format.
        """
        image = ImageParser.to_pillow(image)

        return ColorPicker.get_dominant_color(image, PixelFilterFunction.is_green)

    @staticmethod
    def get_most_common_green_rgb_color_and_similars(image: Union[str, Image.Image, np.ndarray]):
        """
        Returns the most common rgb color and its similar colors
        found in the provided 'image_filename' as a pair of values
        (most_common, similars). Extract them as a pair.
        """
        image = ImageParser.to_pillow(image)

        return ColorPicker.get_dominant_and_similar_colors(image, PixelFilterFunction.is_green, _is_similar_green)

    @staticmethod
    def get_dominant_color(image: Union[str, Image.Image, np.ndarray], pixel_filter_function: PixelFilterFunction = None):
        """
        Opens the provided 'image_filename' and gets the dominant
        color applying the 'pixel_filter_function' if provided.
        """
        image = ImageParser.to_pillow(image)

        dominant, _ = _get_dominant_color(image, pixel_filter_function)

        return dominant

    @staticmethod
    def get_dominant_and_similar_colors(image: Union[str, Image.Image, np.ndarray], pixel_filter_function: PixelFilterFunction = None, similarity_function = None):
        """
        Opens the provided 'image_filename', gets the dominant
        color and also the similar ones by applying the 
        'pixel_filter_function' if provided.
        """
        image = ImageParser.to_pillow(image)

        return _get_dominant_color(image, pixel_filter_function, similarity_function)



def _is_similar_green(color1, color2, tolerance: float = 30):
    if not tolerance:
        tolerance = 30

    # TODO: This below should be comparing 
    return (abs(color1[0] - color2[0]) <= tolerance * 0.5 and
            abs(color1[1] - color2[1]) <= tolerance * 2 and
            abs(color1[2] - color2[2]) <= tolerance * 0.5)

def _get_dominant_color(image: Union[str, Image.Image, np.ndarray], pixel_filter_function: PixelFilterFunction = None, similarity_function = None):
    image = ImageParser.to_pillow(image)
    
    pixels = list(image.getdata())
    if pixel_filter_function is not None:
        pixels = [pixel for pixel in pixels if pixel_filter_function(pixel)]

    color_count = Counter(pixels)

    if not color_count:
        return None, None

    dominant_color = color_count.most_common(1)[0][0] #[0][1] is the 'times'

    if similarity_function is None:
        return dominant_color, None

    similar_colors = [color for color in color_count.keys() if similarity_function(color, dominant_color, 30) and color != dominant_color]
    
    return dominant_color, similar_colors

# TODO: Maybe make this methods work with numpy arrays (?)


from yta_general_utils.web.tiktok.url_parser import parse_tiktok_url
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys

import time


def get_video_metadata_by_url(tiktok_video_url):
    """
    Receives a tiktok video url and build the metadata information by
    scrapping with chromedriver.

    This method returns the video 'id', 'username', 'video_id', 'title'
    and 'description'
    """
    internal_info = parse_tiktok_url(tiktok_video_url)

    try:
        options = Options()
        options.add_argument("--start-maximized")
        # Remove this line below for debug
        #options.add_argument("--headless=new") # for Chrome >= 109
        driver = webdriver.Chrome(options = options)
        driver.get(internal_info['url'])

        time.sleep(10)

        # [ 1 ] Continue as guest
        actions = ActionChains(driver) 
        actions.send_keys(Keys.TAB * 8)
        actions.perform()

        actions = ActionChains(driver)
        actions.send_keys(Keys.ENTER * 1)
        actions.perform()

        time.sleep(5)

        title = driver.title.replace('... | TikTok', '')
        description = ''
        description_elements = driver.find_elements_by_xpath("//*[@data-e2e='browse-video-desc']")[0].find_elements_by_css_selector("*")
        for element in description_elements:
            description += element.get_attribute('innerText') + ' '
        description = description.strip()

    finally:
        driver.close()

        return {
            'title': title,
            'description': description,
            'url': internal_info['url'],
            'username': internal_info['username'],
            'video_id': internal_info['video_id']
        }

def get_popular_tiktok_videos():
    """
    Work in progress.
    """
    # TODO: In progress. To explore most popular videos and build a list of urls

    # Create a Chrome driver
    options = Options()
    options.add_argument("--start-maximized")
    driver = webdriver.Chrome(options = options)

    try:
        # TODO: Improve the way we look for those popular tiktok videos
        # We can download them, but we need to be sure that those videos
        # are actually popular and could be viral
        """
        driver.get(test_url)
        time.sleep(2)
        print(driver.current_url)
        """

        """
        explore_tiktok_url = 'https://www.tiktok.com/explore'
        driver.get(explore_tiktok_url)

        time.sleep(5)

        # [ 1 ] Continue as guest
        actions = ActionChains(driver) 
        actions.send_keys(Keys.TAB * 8)
        actions.perform()

        actions = ActionChains(driver)
        actions.send_keys(Keys.ENTER * 1)
        actions.perform()

        time.sleep(5)

        # [ 2 ] Look for popular videos
        videos = driver.find_elements_by_xpath("//*[@data-e2e='explore-item']")
        
        for video in videos:
            # [ 3 ] Download each of those videos
            video_link_element = video.find_elements_by_css_selector("*")[0].find_elements_by_css_selector("*")[0].find_elements_by_css_selector("*")[0]
            href = video_link_element.get_attribute('href')
            split = href.split('/')
            username = str(split[len(split) - 3]).replace('@', '')
            video_id = split[len(split) - 1]

            # TODO: What about video title?
            output_video_name = username + '_' + video_id + '_tiktok.mp4'
            print('Downloading video as "' + output_video_name + '"')
            download_tiktok_video(href, output_video_name)
        # data-e2e="explore-item" and 3rd son is a href with video link
        """

        """
        If I share a video from mobile Tiktok, the link is like this (https://vm.tiktok.com/ZGeSJ6YRA),
        but if you search that on Chrome, it will be replaced by the whole url like this 
        (https://www.tiktok.com/@ahorayasabesque/video/7327001175616703777?_r=1&_t=8jqq93LWqsC).
        From that long link, if you remove the '?_...' part, you'll get the whole link that we use,
        including the username and the video_id, so we can work propertly with it.

        This could be a way of storing interesting tiktok videos in some places and then
        automate the process of adding some edition to the video (maybe some short video before,
        overlay video, title, etc.) and then upload it to youtube as a short.
        """

    finally:
        driver.close()



from yta_general_utils.color.utils import is_hexadecimal_color, parse_rgba_color, parse_rgb_color
from yta_general_utils.color.converter import ColorConverter
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.color.enums import ColorString
from typing import Union


class Color:
    """
    Class that represents a color, stored as RGBA, makes
    easy the way we interact with color and provide them as
    parameters and simplify the color conversion. The color
    is stored as a not-normalized color, but values can be
    normalized through the methods that allow it (those 
    including the 'normalized' bool parameter).

    Any attribute has to be initialized with a value between
    0 and 255. Alpha as 255 means full opaque.

    TODO: Please confirm alpha 255 is opaque.
    """
    r: int
    """
    Red color, from 0 to 255, where 0 is no value and 255
    is everything.
    """
    g: int
    """
    Green color, from 0 to 255, where 0 is no value and 255
    is everything.
    """
    b: int
    """
    Blue color, from 0 to 255, where 0 is no value and 255
    is everything.
    """
    a: int
    """
    Alpha (transparency), from 0 to 255, where 0 is no
    value and 255 is everything.
    """
    def __init__(self, r, g, b, a):
        self.r, self.g, self.b, self.a = r, g, b, a

    def as_rgb(self, normalized: bool = False):
        """
        Return the color as a tuple of the 3 rgb values
        that are, in order: red, green and blue.
        """
        if normalized:
            return self.r / 255.0, self.g / 255.0, self.b / 255.0
        
        return self.r, self.g, self.b
    
    def as_rgba(self, normalized: bool = False):
        """
        Return the color as a tuple of the 4 rgba values
        that are, in order: red, green, blue and alpha.
        """
        if normalized:
            return self.r / 255.0, self.g / 255.0, self.b / 255.0, self.a / 255.0
        
        return self.r, self.g, self.b, self.a

    def as_rgb_array(self, normalized: bool = False):
        """
        Return the color as an array of the 3 rgb values
        that are, in order: red, green and blue.
        """
        return [*self.as_rgb(normalized)]
    
    def as_rgba_array(self, normalized: bool = False):
        """
        Return the color as an array of the 4 rgba values
        that are, in order: red, green, blue and alpha.
        """
        return [*self.as_rgba(normalized)]
    
    def as_hex(self, do_include_alpha: bool = False):
        """
        Return the color as a string represening it in
        hexadecimal value. The result will be #RRGGBB if
        'do_include_alpha' is False, or #RRGGBBAA if
         True.
        """
        return ColorConverter.rgba_to_hex(self.as_rgba(), do_include_alpha)
    
    def as_hsl(self):
        # TODO: Explain
        return ColorConverter.rgba_to_hsl(self.as_rgba())
    
    def as_cymk(self):
        # TODO: Explain
        return ColorConverter.rgba_to_cymk(self.as_rgba())
    
    # TODO: Use the cv2 library to make other changes
    @staticmethod
    def parse(color: Union[list, tuple, str, 'ColorString', 'Color']):
        """
        Parse the provided 'color' parameter and return the
        color as r,g,b,a values or raises an Exception if it
        is not a valid and parseable color.

        This method accepts string colors (if names are
        registered in our system), hexadecimal colors (than
        can also include alpha value), RGB array or tuples
        (that can be normalized, with float values between
        0.0 and 1.0, or not normalized, with int values
        between 0 and 255), or RGBA array or tuples, similar
        to RGB but including a 4h alpha value.
        """
        if PythonValidator.is_instance(color, Color):
            return color

        string_color = None
        try:
            string_color = ColorString.to_enum(color)
        except:
            pass

        color_array = None
        if string_color is not None:
            color_array = ColorConverter.hex_to_rgba(string_color.value)
        elif PythonValidator.is_string(color) and is_hexadecimal_color(color):
            color_array = ColorConverter.hex_to_rgba(color)
        else:
            try:
                color_array = parse_rgba_color(color)
            except:
                pass

            try:
                color_array = *parse_rgb_color(color), 0
            except:
                pass

            if color_array is None:
                raise Exception(f'The provided "color" parameter is not parseable.')
        
        return Color(*color_array)



from yta_general_utils.programming.regular_expressions import RegularExpression


class ColorRegularExpression(RegularExpression):
    HEX = r'^(#|0x|0X)[0-9a-fA-F]{6,8}$'
    """
    Accepts colors with or without alpha, that include the
    '#', '0x' or '0X' begining.
    """


from yta_general_utils.downloader.utils import download_file
from yta_general_utils.file.filename import filename_is_type, replace_file_extension
from yta_general_utils.file.enums import FileType


def download_audio(url: str, output_filename: str):
    """
    Receives a downloadable url as 'url' and downloads that audio in
    our system as 'output_filename'. The 'output_filename' can be 
    changed to a valid audio filename if it is wrong.

    This method will return the final 'output_filename' in which the
    file has been downloaded if so.
    """
    # TODO: Check if url is a valid audio or not
    if not filename_is_type(output_filename, FileType.AUDIO):
        output_filename = replace_file_extension(output_filename, '.mp3')

    return download_file(url, output_filename)